import numpy as np
import pytest


def test_random_time_stretch():
    """Tests the time_stretch transform with synthetic data for 2 scenarions: slowing down and
    speeding up.

    - verifies if the output of transform is different than the input data
    - verifies if the length of signal has chanched according to the stretch factor (and it should remain fixed if fix_length flag is True)
    """
    from tonic.audio_augmentations import RandomTimeStretch

    np.random.seed(123)

    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl)

    for fix_length in [False, True]:
        # verify length of stretched signal
        slowing_down = RandomTimeStretch(
            samplerate=sr, sample_length=sl, factors=[0.5], fix_length=fix_length
        )
        slow = slowing_down(data)

        assert slow is not data

        if fix_length:
            assert slow.shape[1] == data.shape[1]
        else:
            assert np.allclose(
                slow.shape[1],
                data.shape[1] / (slowing_down.factors[0]),
                rtol=1e-2,
                atol=1e-3,
            )

        speeding_up = RandomTimeStretch(
            samplerate=sr, sample_length=sl, factors=[1.5], fix_length=fix_length
        )
        fast = speeding_up(data)

        assert fast is not data

        if fix_length:
            assert fast.shape[1] == data.shape[1]
        else:
            assert np.allclose(
                fast.shape[1],
                data.shape[1] / (speeding_up.factors[0]),
                rtol=1e-2,
                atol=1e-3,
            )


def test_random_pitch_shift():
    """Tests the pitch_shift transform with synthetic data.

    - verifies if the output of transform is different than the input data
    - verifies that the size has not changed
    """
    from tonic.audio_augmentations import RandomPitchShift

    np.random.seed(123)

    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl)

    aug = RandomPitchShift(samplerate=sr)
    pitch_shifted = aug(data)

    assert pitch_shifted is not data

    assert pitch_shifted.shape[1] == data.shape[1]


def test_random_amplitude_scale():
    """Tests the amplitude_scale transform with synthetic data.

    - verifies if the output of transform is different than the input data
    - verifies that the size has not changed
    - verifies that maximum amplitude is in the defined range
    """
    from tonic.audio_augmentations import RandomAmplitudeScale

    np.random.seed(123)

    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.ones((1, sr * sl))
    min_amp, max_amp = 0.05, 0.15

    aug = RandomAmplitudeScale(samplerate=sr, min_amp=min_amp, max_amp=max_amp)
    amp_scaled = aug(data)

    assert amp_scaled is not data
    assert amp_scaled.shape[1] == data.shape[1]
    assert amp_scaled.max() <= max_amp


def test_add_white_noise():
    """Tests the add_white_noise transform with synthetic data.

    - verifies if the output of transform is different than the input data
    - verifies that the size has not changed
    """
    from tonic.audio_augmentations import AddWhiteNoise

    np.random.seed(123)

    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl)

    aug = AddWhiteNoise(samplerate=sr)
    noisy = aug(data)
    assert noisy is not data
    assert noisy.shape[1] == data.shape[1]


def test_RIR():
    """Tests the RIR transform with a synthetic data.

    - verifies if the output of transform is different than the input data
    - verifies that the size has not changed
    """
    from tonic.audio_augmentations import RIR

    np.random.seed(123)

    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl).astype("float32")
    rir_audio_path = (
        "tutorial-assets/Lab41-SRI-VOiCES-rm1-impulse-mc01-stu-clo-8000hz.wav"
    )
    aug = RIR(samplerate=sr, rir_audio=rir_audio_path)
    RIR_augmented = aug(data)
    assert RIR_augmented is not data
    assert RIR_augmented.shape[1] == data.shape[1]



import os
from typing import Any, Callable, Optional, Tuple

import h5py
import numpy as np

from tonic.dataset import Dataset
from tonic.io import make_structured_array


class ThreeET_Eyetracking(Dataset):
    """3ET DVS eye tracking `3ET <https://github.com/qinche106/cb-convlstm-eyetracking>`_
    ::

        @article{chen20233et,
            title={3ET: Efficient Event-based Eye Tracking using a Change-Based ConvLSTM Network},
            author={Chen, Qinyu and Wang, Zuowen and Liu, Shih-Chii and Gao, Chang},
            journal={arXiv preprint arXiv:2308.11771},
            year={2023}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        transform (callable, optional): A callable of transforms to apply to the data.
        split (string, optional): The dataset split to use, ``train`` or ``val``.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.

    Returns:
         A dataset object that can be indexed or iterated over.
         One sample returns a tuple of (events, targets).
    """

    url = "https://dl.dropboxusercontent.com/s/1hyer8egd8843t9/ThreeET_Eyetracking.zip?dl=0"
    filename = "ThreeET_Eyetracking.zip"
    file_md5 = "b6c652b06fdfd85721f39e2dbe12f4e8"

    sensor_size = (240, 180, 2)
    dtype = np.dtype([("t", int), ("x", int), ("y", int), ("p", int)])
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        split: str = "train",
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        # if not exist, download from url
        if not self._check_exists():
            self.download()

        data_dir = os.path.join(save_to, "ThreeET_Eyetracking")
        # Load filenames from the provided lists
        if split == "train":
            filenames = self.load_filenames(os.path.join(data_dir, "train_files.txt"))
        elif split == "val":
            filenames = self.load_filenames(os.path.join(data_dir, "val_files.txt"))
        else:
            raise ValueError("Invalid split name")

        # Get the data file paths and target file paths
        self.data = [os.path.join(data_dir, "data", f + ".h5") for f in filenames]
        self.targets = [os.path.join(data_dir, "labels", f + ".txt") for f in filenames]

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Returns:
            (events, target) where target is index of the target class.
        """
        # get events from .h5 file
        with h5py.File(self.data[index], "r") as f:
            events = f["events"][:]
        # load the sparse labels
        with open(self.targets[index], "r") as f:
            target = np.array(
                [line.strip().split() for line in f.readlines()], np.float64
            )

        events = make_structured_array(
            events[:, 0],  # time in us
            events[:, 1],  # x
            events[:, 2],  # y
            events[:, 3],  # polarity in 1 or 0
            dtype=self.dtype,
        )

        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return self._is_file_present()

    def load_filenames(self, path):
        with open(path, "r") as f:
            return [line.strip() for line in f.readlines()]



import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured

from tonic.functional.to_frame import to_frame_numpy

def differentiator_downsample(events: np.ndarray, sensor_size: tuple, target_size: tuple, dt: float, 
                              differentiator_time_bins: int = 2, noise_threshold: int = 0):
    """Spatio-temporally downsample using the integrator method coupled with a differentiator to effectively 
    downsample large object sizes relative to downsampled pixel resolution in the DVS camera's visual field.
    
    Incorporates the paper Ghosh et al. 2023, Insect-inspired Spatio-temporal Downsampling of Event-based Input,
    https://doi.org/10.1145/3589737.3605994
    
    Parameters:
        events (ndarray): ndarray of shape [num_events, num_event_channels].
        sensor_size (tuple): a 3-tuple of x,y,p for sensor_size.
        target_size (tuple): a 2-tuple of x,y denoting new down-sampled size for events to be
                             re-scaled to (new_width, new_height).
        dt (float): step size for simulation, in ms.
        differentiator_time_bins (int): number of equally spaced time bins with respect to the dt 
                                        to be used for the differentiator.
        noise_threshold (int): number of events before a spike representing a new event is emitted.
        
    Returns:
        the spatio-temporally downsampled input events using the differentiator method.
    """
        
    assert "x" and "y" and "t" in events.dtype.names
    assert np.logical_and(np.remainder(differentiator_time_bins, 1) == 0, differentiator_time_bins >= 1)
    
    events = events.copy()
    
    # Call integrator method
    dt_scaling, events_integrated = integrator_downsample(events, sensor_size=sensor_size, target_size=target_size, 
                                                          dt=(dt / differentiator_time_bins), 
                                                          noise_threshold=noise_threshold, differentiator_call=True)
    
    if dt_scaling:
        dt *= 1000
        
    num_frames = int(events_integrated[-1][0] // dt + 1)
    frame_histogram = np.zeros((num_frames, *np.flip(target_size), 2))
        
    for event in events_integrated:
        differentiated_time, event_histogram = event
        time = int(differentiated_time // dt)
        
        # Separate events based on polarity and apply Heaviside
        event_hist_pos = event_histogram >= noise_threshold
        event_hist_neg = -event_histogram >= noise_threshold
        
        frame_histogram[time,...,1] += event_hist_pos
        frame_histogram[time,...,0] += event_hist_neg
        
    # Differences between subsequent frames
    frame_differences = (np.diff(frame_histogram, axis=0)).clip(min=0)
    
    # Restructuring numpy array to structured array
    time_index, y_new, x_new, polarity_new = np.nonzero(frame_differences)
    
    events_new = np.column_stack((x_new, y_new, polarity_new.astype(dtype=bool), time_index * dt))
    
    names = ["x", "y", "p", "t"]
    formats = ['i4', 'i4', 'i4', 'i4']
    
    dtype = np.dtype({'names': names, 'formats': formats})
    
    return unstructured_to_structured(events_new.copy(), dtype=dtype)
    
def integrator_downsample(events: np.ndarray, sensor_size: tuple, target_size: tuple, dt: float, noise_threshold: int = 0, 
                          differentiator_call: bool = False):
    """Spatio-temporally downsample using with the following steps:
    
    1. Differencing of ON and OFF events to counter camera shake or jerk.
    2. Use an integrate-and-fire (I-F) neuron model with a noise threshold similar to 
    the membrane potential threshold in the I-F model to eliminate high-frequency noise.
    
    Multiply x/y values by a spatial_factor obtained by dividing sensor size by the target size.
    
    Parameters:
        events (ndarray): ndarray of shape [num_events, num_event_channels].
        sensor_size (tuple): a 3-tuple of x,y,p for sensor_size.
        target_size (tuple): a 2-tuple of x,y denoting new down-sampled size for events to be
                             re-scaled to (new_width, new_height).
        dt (float): temporal resolution of events in milliseconds.
        noise_threshold (int): number of events before a spike representing a new event is emitted.
        differentiator_call (bool): Preserve frame spikes for differentiator method in order to optimise 
                                    differentiator method.
        
    Returns:
        the spatio-temporally downsampled input events using the integrator method.
    """
    
    assert "x" and "y" and "t" in events.dtype.names
    assert isinstance(noise_threshold, int)
    assert dt is not None
    
    events = events.copy()
    
    if np.issubdtype(events["t"].dtype, np.integer):
        dt *= 1000
        dt_scaling = True
    
    if differentiator_call:
        assert dt // events["t"][-1] == 0
    
    # Downsample
    spatial_factor = np.asarray(target_size) / sensor_size[:-1]

    events["x"] = events["x"] * spatial_factor[0]
    events["y"] = events["y"] * spatial_factor[1]
    
    # Compute all histograms at once
    all_frame_histograms = to_frame_numpy(events, sensor_size=(*target_size, 2), time_window=dt)
    
    # Subtract the channels for ON/OFF differencing
    frame_histogram_diffs = all_frame_histograms[:, 1] - all_frame_histograms[:, 0]
    
    frame_spike = np.zeros(np.flip(target_size))
    event_histogram = []
    
    events_new = []
    
    for time, frame_histogram in enumerate(frame_histogram_diffs):
    
        frame_spike += frame_histogram
            
        coordinates_pos = np.stack(np.nonzero(np.maximum(frame_spike >= noise_threshold, 0))).T
        coordinates_neg = np.stack(np.nonzero(np.maximum(-frame_spike >= noise_threshold, 0))).T
        
        if np.logical_or(coordinates_pos.size, coordinates_neg.size).sum():
        
            # For optimising differentiator
            event_histogram.append((time*dt, frame_spike.copy()))
            
            # Reset spiking coordinates to zero
            frame_spike[coordinates_pos[:,0], coordinates_pos[:,1]] = 0
            frame_spike[coordinates_neg[:,0], coordinates_neg[:,1]] = 0
            
            # Restructure events
            events_new.append(np.column_stack((np.flip(coordinates_pos, axis=1), np.ones((coordinates_pos.shape[0],1)).astype(dtype=bool), 
                                                (time*dt)*np.ones((coordinates_pos.shape[0],1)))))
            
            events_new.append(np.column_stack((np.flip(coordinates_neg, axis=1), np.zeros((coordinates_neg.shape[0],1)).astype(dtype=bool), 
                                                (time*dt)*np.ones((coordinates_neg.shape[0],1)))))
        
    if differentiator_call:
        return dt_scaling, event_histogram
    else:
        events_new = np.concatenate(events_new.copy())
        
        names = ["x", "y", "p", "t"]
        formats = ['i4', 'i4', 'i4', 'i4']
        
        dtype = np.dtype({'names': names, 'formats': formats})
        
        return unstructured_to_structured(events_new.copy(), dtype=dtype)



import numpy as np
import pytest


def test_standardize_data_length():
    from tonic.audio_transforms import FixLength

    sdl = FixLength(100, 1)

    # Data is longer
    data = np.ones((1, 120))
    assert sdl(data).shape == (1, 100)

    # Data is shorter
    data = np.ones((1, 80))
    assert sdl(data).shape == (1, 100)


def test_bin():
    import numpy as np

    from tonic.audio_transforms import Bin

    bin_transform = Bin(orig_freq=16_000, new_freq=100, axis=1)

    data = np.random.random((1, 8 * 16_000))
    data_binned = bin_transform(data)

    assert data_binned.shape == (1, 8 * 100)
    assert data.sum() == pytest.approx(data_binned.sum(), 1e-7)


def test_linear_butter_filter_bank():
    import numpy as np

    from tonic.audio_transforms import LinearButterFilterBank

    fb = LinearButterFilterBank(
        order=2, low_freq=100, sampling_freq=16000, num_filters=16
    )
    data = np.random.random((1, 16_000))

    filter_out = fb(data)
    assert filter_out.shape == (16, 16_000)


def test_mel_butter_filter_bank():
    import numpy as np

    from tonic.audio_transforms import MelButterFilterBank

    fb = MelButterFilterBank(order=2, low_freq=100, sampling_freq=16000, num_filters=16)
    data = np.random.random((1, 16_000))

    filter_out = fb(data)
    assert filter_out.shape == (16, 16_000)


def test_add_noise():
    import numpy as np

    class DummyNoiseDataset:
        def __len__(self):
            return 1000

        def __getitem__(self, item):
            sig_len = np.random.randint(12000, 20000, (1,)).item()
            return np.random.random((1, sig_len)), 0

    data = np.sin(np.arange(0, 16_000 * 0.001, 0.001))[None, ...]
    print(data.shape)

    from tonic.audio_transforms import AddNoise

    noise_dataset = DummyNoiseDataset()

    print(noise_dataset[0][0].shape)

    add_noise = AddNoise(noise_dataset, 10, normed=True)

    signal = add_noise(data)
    assert signal.shape == (1, 16_000)


def test_swap_axes():
    """Tests SwapAxes transform with synthetic data."""
    from tonic.audio_transforms import SwapAxes

    np.random.seed(123)
    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl)
    ax1, ax2 = 0, 1
    swap_ax = SwapAxes(ax1=ax1, ax2=ax2)
    swaped = swap_ax(data)

    assert swaped.shape[0] == data.shape[1]
    assert swaped.shape[1] == data.shape[0]


def test_amplitude_scale():
    """Tests the amplitude scaling transform with synthetic data."""
    from tonic.audio_transforms import AmplitudeScale

    np.random.seed(123)
    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl)
    max_amps = np.random.rand(10)

    for amp in max_amps:
        AmpScale = AmplitudeScale(max_amplitude=amp)
        transformed = AmpScale(data)
        assert data.shape[1] == transformed.shape[1]
        assert transformed.max() == amp


def test_robust_amplitude_scale():
    """Tests robust amplitude scaling transform with a synthetic data."""
    from tonic.audio_transforms import RobustAmplitudeScale

    np.random.seed(123)
    sr = 16_000  # sample rate
    sl = 1  # sample length
    data = np.random.rand(1, sr * sl)
    max_amps = np.random.rand(10)
    percent = 0.01
    for amp in max_amps:
        RobustAmpScale = RobustAmplitudeScale(
            max_robust_amplitude=amp, outlier_percent=percent
        )
        transformed = RobustAmpScale(data)
        sorted_transformed = np.sort(np.abs(transformed.ravel()))
        non_outlier = sorted_transformed[
            0 : int(np.floor(len(sorted_transformed)) * (1 - percent))
        ]
        print(non_outlier)
        assert data.shape[1] == transformed.shape[1]
        assert np.all(non_outlier <= amp)



import os
import numpy as np
from typing import Any, Callable, Optional, Tuple
from tonic.download_utils import extract_archive
from tonic.io import make_structured_array

class NERDD:
    """`NeRDD <https://github.com/MagriniGabriele/NeRDD>`_
    Neuromorphic Event-based Reactive Driving Dataset.

    Parameters:
        root (string): Location to save files to on disk.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    filename = "NeRDD.zip"
    folder_name = "DATA"
    sensor_size = (1280, 720, 2)
    dtype = np.dtype([("t", np.int32), ("x", np.int16), ("y", np.int16), ("p", bool)])
    ordering = dtype.names

    def __init__(
        self,
        root: str,
        transforms: Optional[Callable] = None,
    ):
        self.data = []
        self.targets = []
        self.location_on_system = root
        self.transforms = transforms

        if not self._check_extracted():
            if not self._check_zip_exists():
                raise FileNotFoundError(
                    f"The dataset archive '{self.filename}' was not found in {self.location_on_system}. "
                    "Please download it manually and place it in the specified directory."
                )
            self._extract_archive()
        
        self._load_dataset_structure()

    def _check_zip_exists(self) -> bool:
        """Check if the zip file exists in the specified directory."""
        return os.path.isfile(os.path.join(self.location_on_system, self.filename))

    def _check_extracted(self) -> bool:
        """Check if the dataset has been extracted and contains at least 10 .npz files."""
        data_path = os.path.join(self.location_on_system, self.folder_name)
        if not os.path.isdir(data_path):
            return False
        npz_files = []
        for root, _, files in os.walk(data_path):
            npz_files.extend([f for f in files if f.endswith(".npz")])
        return len(npz_files) == 115

    def _extract_archive(self):
        """Extract the dataset archive."""
        print(f"Extracting {self.filename}...")
        archive_path = os.path.join(self.location_on_system, self.filename)
        extract_archive(archive_path)
        print(f"Extraction complete. Files are now in {self.location_on_system}.")

    def _load_dataset_structure(self):
        """Load the dataset files and their corresponding labels."""
        data_path = os.path.join(self.location_on_system, self.folder_name)
        archives = [
            f for f in os.listdir(data_path) 
            if os.path.isdir(os.path.join(data_path, f))
        ]
        
        for archive in archives:
            archive_path = os.path.join(data_path, archive)
            scenes = [
                f for f in os.listdir(archive_path) 
                if os.path.isdir(os.path.join(archive_path, f))
            ]
            
            for scene in scenes:
                scene_path = os.path.join(archive_path, scene)
                event_file = os.path.join(scene_path, "Event", "output_events.npz")
                rgb_folder = os.path.join(scene_path, "rgb_folder")
                label_file = os.path.join(scene_path, "ev_rgb_coordinates.txt")
                
                if os.path.exists(event_file) and os.path.exists(label_file):
                    self.data.append((event_file, label_file))
                else:
                    print(f"Skipping scene {scene} in archive {archive}: missing files.")

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Returns:
            (events, target) where target is index of the target class.
        """
        event_file, label_file = self.data[index]
        events = np.load(event_file)["data"]
        events = make_structured_array(
            events[:, 3],  # t
            events[:, 0],  # x
            events[:, 1],  # y
            events[:, 2],  # p
            dtype=self.dtype,
        )

        with open(label_file, "r") as f:
            bboxes = []
            for line in f.readlines():
                timestamp, bbox_str = line.strip().split(":")
                bbox_coords = list(map(float, bbox_str.split(",")))
                bboxes.append([float(timestamp)] + bbox_coords)
            bboxes = np.array(bboxes, dtype=np.float32)

        # Apply transforms if provided
        if self.transforms is not None:
            events, bboxes = self.transforms(events, bboxes)

        return events, bboxes

    def __len__(self):
        return len(self.data)


import os

import numpy as np

from tonic.audio_augmentations import RandomPitchShift
from tonic.audio_transforms import AmplitudeScale, FixLength
from tonic.cached_dataset import Aug_DiskCachedDataset, load_from_disk_cache


class mini_dataset:
    def __init__(self) -> None:
        np.random.seed(0)
        self.data = np.random.rand(10, 16000)
        self.transform = None
        self.target_transform = None

    def __getitem__(self, index):
        sample = self.data[index]
        label = 1
        if sample.ndim == 1:
            sample = sample[None, ...]
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            label = self.target_transform(label)

        return sample, label


def test_aug_disk_caching():
    from torchvision.transforms import Compose

    all_transforms = {}
    all_transforms["pre_aug"] = [AmplitudeScale(max_amplitude=0.150)]
    all_transforms["augmentations"] = [RandomPitchShift(samplerate=16000, caching=True)]
    all_transforms["post_aug"] = [FixLength(16000)]
    # number of copies is set to number of augmentation params (factors)
    n = len(RandomPitchShift(samplerate=16000, caching=True).factors)
    Aug_cach = Aug_DiskCachedDataset(
        dataset=mini_dataset(),
        cache_path="cache/",
        all_transforms=all_transforms,
        num_copies=n,
    )

    if not os.path.isdir("cache/"):
        os.mkdir("cache/")

    sample_index = 0
    Aug_cach.generate_all(sample_index)

    for i in range(n):
        transform = Compose(
            [
                AmplitudeScale(max_amplitude=0.150),
                RandomPitchShift(samplerate=16000, caching=True, aug_index=i),
                FixLength(16000),
            ]
        )
        ds = mini_dataset()
        ds.transform = transform
        augmented_sample = ds[sample_index][0]
        loaded_sample, targets = load_from_disk_cache(
            "cache/" + "0_" + str(i) + ".hdf5"
        )
        assert (augmented_sample == loaded_sample).all()



#!/user/bin/env python

import numpy as np
import h5py
import os
from typing import Callable, Optional

from tonic.dataset import Dataset
from tonic.io import make_structured_array
import requests
from tqdm import tqdm

class NTIDIGITS18(Dataset):
    """`N-TIDIGITS18 Dataset <https://docs.google.com/document/d/1Uxe7GsKKXcy6SlDUX4hoJVAC0-UkH-8kr5UXp0Ndi1M/edit?tab=t.0#heading=h.sbnu5gtazqjq/>`_
    Cochlea Spike Dataset.
    ::

        @article{anumula2018feature,
          title={Feature representations for neuromorphic audio spike streams},
          author={Anumula, Jithendar and Neil, Daniel and Delbruck, Tobi and Liu, Shih-Chii},
          journal={Frontiers in neuroscience},
          volume={12},
          pages={23},
          year={2018},
          publisher={Frontiers Media SA}
        }

    Parameters:
        save_to (string): Location to save files to on disk. Will put files in an 'hsd' subfolder.
        train (bool): If True, uses training subset, otherwise testing subset.
        single_digits (bool): If True, only returns samples with single digits (o, 1, 2, 3, 4, 5, 6, 7, 8, 9, z), with class 0 for 'o' and 11 for 'z'.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.

    Returns:
        A dataset object that can be indexed or iterated over. One sample returns a tuple of (events, targets).
    """

    base_url = "https://www.dropbox.com/scl/fi/1x4lxt9yyw25sc3tez8oi/n-tidigits.hdf5?e=2&rlkey=w8gi5udvib2zqzosusa5tr3wq&dl=1"
    filename = "n-tidigits.hdf5"
    file_md5 = "360a2d11e5429555c9197381cf6b58e0"
    folder_name = ""

    sensor_size = (64, 1, 1)
    dtype = np.dtype([("t", int), ("x", int), ("p", int)])
    ordering = dtype.names

    class_map = {"o": 0,
                 "1": 1,
                 "2": 2,
                 "3": 3,
                 "4": 4,
                 "5": 5,
                 "6": 6,
                 "7": 7,
                 "8": 8,
                 "9": 9,
                 "z": 10}

    def __init__(
            self,
            save_to: str,
            train: bool = True,
            single_digits=False,
            transform: Optional[Callable] = None,
            target_transform: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
        )

        self.url = self.base_url
        # load the data
        self.file_path = os.path.join(self.location_on_system, self.filename)

        if not self._check_exists():
            self.download()

        self.data = h5py.File(self.file_path, 'r')
        self.partition = "train" if train else "test"
        self.single_indices = [i for i in range(len(self.data[f"{self.partition}_labels"])) if
                               len(self.data[f"{self.partition}_labels"][i].decode().split("-")[-1]) == 1]
        self._samples = [x.decode() for x in self.data[f"{self.partition}_labels"]]
        self.single_digits = single_digits

        if single_digits:
            self._samples = [self._samples[i] for i in self.single_indices]

        self.labels = [x.decode().split("-")[-1] for x in self.data[f"{self.partition}_labels"]]

    def download(self) -> None:
        response = requests.get(self.base_url, stream=True)
        if response.status_code == 200:
            print("Downloading N-TIDIGITS from Dropbox at {}...".format(self.base_url))
            file_size = int(response.headers.get('Content-Length', 0))  # get total file size in bytes
            chunk_size = 8192

            os.makedirs(self.location_on_system, exist_ok=True)
            # Initialize progress bar
            with open(os.path.join(self.location_on_system, self.filename), 'wb') as f, tqdm(
                    total=file_size,
                    unit='B',
                    unit_scale=True,
                    desc="Downloading",
                    ascii=True
            ) as pbar:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        pbar.update(len(chunk))
        else:
            print("Failed to download N-TIDIGITS from Dropbox. Please try again later.")
            response.raise_for_status()
    def __getitem__(self, index):
        sample_id = self._samples[index]
        x = np.asarray(self.data[f"{self.partition}_addresses"][sample_id])
        t = np.asarray(self.data[f"{self.partition}_timestamps"][sample_id])
        events = make_structured_array(
            t * 1e6,
            x,
            1,
            dtype=self.dtype,
        )

        target = sample_id.split("-")[-1]

        if self.single_digits:
            assert len(target) == 1, "Single digit samples requested, but target is not single digit."
            target = self.class_map[target]

        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self._samples)

    def _check_exists(self):
        return (
            self._is_file_present()
        )


from .asl_dvs import ASLDVS
from .cifar10dvs import CIFAR10DVS
from .davisdataset import DAVISDATA
from .dsec import DSEC
from .dvs_lips import DVSLip
from .dvsgesture import DVSGesture
from .ebssa import EBSSA
from .ntidigits18 import NTIDIGITS18
from .hsd import SHD, SSC
from .mvsec import MVSEC
from .ncaltech101 import NCALTECH101
from .nerdd import NERDD
from .nmnist import NMNIST
from .pokerdvs import POKERDVS
from .s_mnist import SMNIST
from .threeET_eyetracking import ThreeET_Eyetracking
from .tum_vie import TUMVIE
from .visual_place_recognition import VPR

__all__ = [
    "ASLDVS",
    "CIFAR10DVS",
    "DAVISDATA",
    "DSEC",
    "DVSGesture",
    "EBSSA",
    "MVSEC",
    "NCALTECH101",
    "NERDD",
    "NMNIST",
    "POKERDVS",
    "SHD",
    "SMNIST",
    "SSC",
    "ThreeET_Eyetracking",
    "TUMVIE",
    "VPR",
    "DVSLip",
]



"""
=================
EventDownsampling
=================
The :class:`~tonic.transforms.EventDownsampling` applies 
spatio-temporal downsampling to events as per the downsampling method chosen.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

transform = tonic.transforms.Compose(
    [
        tonic.transforms.EventDownsampling(sensor_size=nmnist.sensor_size, 
                                           target_size=(12, 12), 
                                           dt=0.01, 
                                           downsampling_method="differentiator", 
                                           noise_threshold=0, 
                                           differentiator_time_bins=2),
        tonic.transforms.ToFrame(
            sensor_size=(12, 12, 2),
            time_window=10000,
        ),
    ]
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import os

import dataset_utils
import h5py
import numpy as np
from utils import create_random_input

import tonic.datasets as datasets
from tonic.download_utils import download_url

base_url = "https://nextcloud.lenzgregor.com/s/"


class ASLDVSTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.ASLDVS
    FEATURE_TYPES = (datasets.ASLDVS.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "ASLDVS/a")
        os.makedirs(testfolder, exist_ok=True)
        filename = "2aeALcfARAS8Dkf/download/a_0244.mat"
        download_url(url=base_url + filename, root=testfolder, filename="a_0244.mat")
        return {"n_samples": 1}


class DVSGestureTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSGesture
    FEATURE_TYPES = (datasets.DVSGesture.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSGesture/ibmGestureTrain/user24_led")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(dtype=datasets.DVSGesture.dtype)
        events = np.lib.recfunctions.structured_to_unstructured(events)
        np.save(testfolder + "/0.npy", events)
        np.save(testfolder + "/1.npy", events)
        return {"n_samples": 2}


class DVSGestureTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSGesture
    FEATURE_TYPES = (datasets.DVSGesture.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSGesture/ibmGestureTest/user24_led")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(dtype=datasets.DVSGesture.dtype)
        events = np.lib.recfunctions.structured_to_unstructured(events)
        np.save(testfolder + "/0.npy", events)
        return {"n_samples": 1}


class DVSLipTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSLip
    FEATURE_TYPES = (datasets.DVSLip.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSLip/DVS-Lip/train/accused")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(
            dtype=np.dtype([("t", "<i4"), ("x", "i1"), ("y", "i1"), ("p", "i1")])
        )
        np.save(testfolder + "/0.npy", events)
        np.save(testfolder + "/1.npy", events)
        return {"n_samples": 2}


class DVSLipTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.DVSLip
    FEATURE_TYPES = (datasets.DVSLip.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "DVSLip/DVS-Lip/test/accused")
        os.makedirs(testfolder, exist_ok=True)
        events, sensor_size = create_random_input(
            dtype=np.dtype([("t", "<i4"), ("x", "i1"), ("y", "i1"), ("p", "i1")])
        )
        np.save(testfolder + "/0.npy", events)
        return {"n_samples": 1}


class EBSSATestCase(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.EBSSA
    FEATURE_TYPES = (datasets.EBSSA.dtype,)
    TARGET_TYPES = (np.ndarray,)
    KWARGS = {"split": "labelled"}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "EBSSA")
        os.makedirs(testfolder, exist_ok=True)
        filename = "Jpw3Adae5kReMrN/download/labelled_ebssa.h5"
        download_url(
            url=base_url + filename, root=testfolder, filename="labelled_ebssa.h5"
        )
        return {"n_samples": 1}


class ThreeET_EyetrackingTestCase(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.ThreeET_Eyetracking
    FEATURE_TYPES = (datasets.ThreeET_Eyetracking.dtype,)
    TARGET_TYPES = (np.ndarray,)
    KWARGS = {"split": "train"}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "ThreeET_Eyetracking")
        os.makedirs(testfolder, exist_ok=True)
        os.makedirs(os.path.join(testfolder, "data"), exist_ok=True)
        os.makedirs(os.path.join(testfolder, "labels"), exist_ok=True)
        # write one line of file name into train_files.txt under testfolder
        os.system("echo testcase > " + os.path.join(testfolder, "train_files.txt"))
        filename = "testcase"

        # download test h5 file
        download_url(
            url=base_url + "4aiA4BAqz5km4Gc/download/" + filename + ".h5",
            root=os.path.join(testfolder, "data"),
            filename=filename + ".h5",
        )
        # # download test labels
        download_url(
            url=base_url + "G6ejNmXNnB2sKyc/download/" + filename + ".txt",
            root=os.path.join(testfolder, "labels"),
            filename=filename + ".txt",
        )

        return {"n_samples": 1}


class NCaltech101TestCase(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NCALTECH101
    FEATURE_TYPES = (datasets.NCALTECH101.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NCALTECH101/Caltech101/airplanes/")
        os.makedirs(testfolder, exist_ok=True)
        filename = "image_0006.bin"
        url = base_url + "sGTckK5fgit7QH3/download/" + filename
        download_url(url=url, root=testfolder, filename=filename)
        return {"n_samples": 1}


class NMNISTTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NMNIST/Train/1/")
        os.makedirs(testfolder, exist_ok=True)
        filename = "00015.bin"
        url = base_url + "pi6WkPbg6tgd7ca/download/" + filename
        download_url(url=url, root=testfolder, filename=filename)
        return {"n_samples": 1}


class NMNISTTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NMNIST
    FEATURE_TYPES = (datasets.NMNIST.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NMNIST/Test/1/")
        os.makedirs(testfolder, exist_ok=True)
        filename = "00015.bin"
        url = base_url + "pi6WkPbg6tgd7ca/download/" + filename
        download_url(url=url, root=testfolder, filename=filename)
        return {"n_samples": 1}


def create_ntidigits_data(filename, n_samples):
    with h5py.File(filename, mode="w") as write_file:
        # Generate random times and units
        times = np.random.random(size=(n_samples, 100)).astype(np.float16)
        units = np.random.randint(0, 64, size=times.shape, dtype=np.uint16)

        # Generate random sequences of symbols
        symbols = ["o", "1", "2", "3", "4", "5", "6", "7", "8", "9", "z"]
        sequences = np.array([
            "speaker-a-" + "".join(np.random.choice(symbols, size=np.random.randint(1, 7), replace=True))
            for _ in range(n_samples)
        ]).astype("|S18")

        for partition in ["train", "test"]:
            # Create a dictionary to store the addresses and timestamps of each speaker
            addresses = {}
            timestamps = {}
            for i in range(n_samples):
                speaker = sequences[i]
                if speaker not in addresses:
                    addresses[speaker] = units[i]
                    timestamps[speaker] = times[i]

            # Create a group for addresses and timestamps, Store each speaker's data as a dataset in the group
            train_address_group = write_file.create_group("{}_addresses".format(partition))
            for speaker, addresses in addresses.items():
                train_address_group.create_dataset(speaker, data=np.array(addresses, dtype=np.uint16))

            train_timestamps_group = write_file.create_group("{}_timestamps".format(partition))
            for speaker, timestamps in timestamps.items():
                train_timestamps_group.create_dataset(speaker, data=np.array(timestamps, dtype=np.float16))

            # Create datasets for labels
            write_file.create_dataset("{}_labels".format(partition), data=sequences)


class NTIDIGITS18TestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NTIDIGITS18
    FEATURE_TYPES = (datasets.NTIDIGITS18.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NTIDIGITS18/")
        os.makedirs(testfolder, exist_ok=True)
        create_ntidigits_data(testfolder + "n-tidigits.hdf5", n_samples=2)
        return {"n_samples": 2}


class NTIDIGITS18TestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.NTIDIGITS18
    FEATURE_TYPES = (datasets.NTIDIGITS18.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "NTIDIGITS18/")
        os.makedirs(testfolder, exist_ok=True)
        create_ntidigits_data(testfolder + "n-tidigits.hdf5", n_samples=2)
        return {"n_samples": 2}


def create_hsd_data(filename, n_samples):
    with h5py.File(filename, mode="w") as write_file:
        times = np.random.random(size=(n_samples, 100)).astype(np.float16)
        units = (np.random.random(size=(n_samples, 100)) * 700).astype(np.uint16)
        keys = ["zero", "one"]
        speaker = (np.random.random(size=n_samples) * 20).astype(np.uint16)
        write_file.create_dataset("spikes/units", data=units)
        write_file.create_dataset("spikes/times", data=times)
        write_file.create_dataset("labels", data=[1] * n_samples)
        write_file.create_dataset("extra/keys", data=keys)
        write_file.create_dataset("extra/speaker", data=speaker)


class SHDTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.SHD
    FEATURE_TYPES = (datasets.SHD.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": True}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "SHD/")
        os.makedirs(testfolder, exist_ok=True)
        create_hsd_data(testfolder + "shd_train.h5", n_samples=2)
        return {"n_samples": 2}


class SHDTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.SHD
    FEATURE_TYPES = (datasets.SHD.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"train": False}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "SHD/")
        os.makedirs(testfolder, exist_ok=True)
        create_hsd_data(testfolder + "shd_test.h5", n_samples=1)
        return {"n_samples": 1}


class SSCTestCaseTrain(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.SSC
    FEATURE_TYPES = (datasets.SSC.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"split": "train"}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "SSC/")
        os.makedirs(testfolder, exist_ok=True)
        create_hsd_data(testfolder + "ssc_train.h5", n_samples=3)
        return {"n_samples": 3}


class SSCTestCaseValid(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.SSC
    FEATURE_TYPES = (datasets.SSC.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"split": "valid"}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "SSC/")
        os.makedirs(testfolder, exist_ok=True)
        create_hsd_data(testfolder + "ssc_valid.h5", n_samples=4)
        return {"n_samples": 4}


class SSCTestCaseTest(dataset_utils.DatasetTestCase):
    DATASET_CLASS = datasets.SSC
    FEATURE_TYPES = (datasets.SSC.dtype,)
    TARGET_TYPES = (int,)
    KWARGS = {"split": "test"}

    def inject_fake_data(self, tmpdir):
        testfolder = os.path.join(tmpdir, "SSC/")
        os.makedirs(testfolder, exist_ok=True)
        create_hsd_data(testfolder + "ssc_test.h5", n_samples=5)
        return {"n_samples": 5}



from pathlib import Path
from typing import Any, Callable, Optional, Tuple

import h5py
import numpy as np

from tonic.dataset import Dataset
from tonic.io import events_struct, make_structured_array


class EBSSA(Dataset):
    """`EBSSA <https://www.westernsydney.edu.au/icns/resources/reproducible_research3/publication_support_materials2/space_imaging>`_

    There are six different splits provided in this dataset. The labelled section of the dataset contains 84 recordings and 84 label files. 
    The unlabelled section of the dataset contains 153 recordings in folders marked "Unlabelled".
    If the automatic download from Google Drive fails, please download the file manually from https://drive.google.com/uc?id=1lCh2HWvxEzzaBHT5TlPuyUn6XPM5OVWN
    and put it in a folder called 'EBSSA'. Then point the path to its parent folder. For example, if you put the data file in 'datasets/EBSSA/labelled_ebssa.h5',
    create the dataset object with 'dataset = tonic.datasets.EBSSA('datasets', split='labelled')'

    ::

        @article{afshar2020event,
            title={Event-based object detection and tracking for space situational awareness},
            author={Afshar, Saeed and Nicholson, Andrew Peter and Van Schaik, Andre and Cohen, Gregory},
            journal={IEEE Sensors Journal},
            volume={20},
            number={24},
            pages={15117--15132},
            year={2020},
            publisher={IEEE}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        split (string): Which split to load. One of "labelled", "unlabelled", "all".
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    # These are Google drive file IDs found by right clicking on the file and selecting "Get shareable link"
    file_id = "1lCh2HWvxEzzaBHT5TlPuyUn6XPM5OVWN"
    folder_name = ""
    file_name = "labelled_ebssa.h5"
    sensor_size = (240, 180, 2)
    dtype = events_struct
    ordering = dtype.names

    def __init__(
        self,
        save_to: str,
        split: str = "labelled",
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )
        self.split = split
        if split != "labelled":
            raise NotImplementedError("Only labelled split is supported at the moment")

        if not self._check_exists():
            self.download()

        file = h5py.File(Path(self.location_on_system) / self.file_name, "r")
        self.keys = list(file.keys())

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Returns:
            (events, target) where target is dict of bounding box and recording id.
        """
        file = h5py.File(Path(self.location_on_system) / self.file_name, "r")
        data = file[self.keys[index]]

        td = data["TD"]
        t = td["ts"][()].flatten()
        x = td["x"][()].flatten() - 1
        y = td["y"][()].flatten() - 1
        p = td["p"][()].flatten()
        events = make_structured_array(x, y, t, p)
        sensor_x_max, sensor_y_max = events["x"].max() + 1, events["y"].max() + 1
        start_ts = float(events["t"][0])
        events["t"] = events["t"] - start_ts

        if "Obj" in list(data.keys()):
            obj = data["Obj"]
            annotation_time_window = 10_000
            bb_width = 10

            obj_t = obj["ts"][()].flatten() - start_ts
            obj_t = (obj_t / annotation_time_window).round() * annotation_time_window
            x_min = obj["x"][()].flatten() - 1 - bb_width / 2
            x_min = x_min.clip(min=0)
            y_min = obj["y"][()].flatten() - 1 - bb_width / 2
            y_min = y_min.clip(min=0)
            obj_id = obj["id"][()].flatten()

            x_max = x_min + bb_width
            x_max = x_max.clip(max=sensor_x_max - 1)
            y_max = y_min + bb_width
            y_max = y_max.clip(max=sensor_y_max - 1)

            dtype = np.dtype(
                [
                    ("t", np.int64),
                    ("x_min", float),
                    ("y_min", float),
                    ("x_max", float),
                    ("y_max", float),
                    ("id", np.uint8),
                ]
            )
            bbox = make_structured_array(
                obj_t, x_min, y_min, x_max, y_max, obj_id, dtype=dtype
            )
        else:
            bbox = None

        target = {
            "bbox": bbox,
            "recording_id": self.keys[index],
        }

        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.transforms is not None:
            events, target = self.transforms(events, target)
        return events, target

    def __len__(self):
        return len(self.keys)

    def _check_exists(self):
        return self._folder_contains_at_least_n_files_of_type(1, ".h5")

    def download(self):
        import gdown

        Path(self.location_on_system).mkdir(exist_ok=True)
        id = self.file_id
        if not Path(self.location_on_system, id).exists():
            gdown.download(id=id, quiet=False)



import random
from dataclasses import dataclass, field

import librosa
import numpy as np
import torch
import torchaudio
from torchaudio.utils import download_asset

from tonic.audio_transforms import FixLength

__all__ = [
    "RandomTimeStretch",
    "RandomPitchShift",
    "RandomAmplitudeScale",
    "AddWhiteNoise",
    "RIR",
]


@dataclass
class RandomTimeStretch:
    """Time-stretch an audio sample by a fixed rate.
    Parameters:
        samplerate (float): sample rate of the sample
        sample_length (int): sample length in seconds
        factors (float): range of desired factors for time stretch
        aug_index (int): index of the chosen factor for time stretch. It will be randomly chosen from the desired range (if not passed while initilization)
        caching (bool): if we are caching the DiskCached dataset will dynamically pass copy index of data item to the transform (to set aug_index). Otherwise the aug_index will be chosen randomly in every call of transform
        fix_length (bool): if True, time stretched signal will be returned in a fixed length (samplerate * sample_length )
    Args:
        audio (np.ndarray): data sample
    Returns:
        np.ndarray: time stretched data sample
    """

    samplerate: float
    sample_length: int
    factors: list = field(default_factory=lambda: [0.8, 0.9, 1.1, 1.2])
    aug_index: int = 0
    caching: bool = False
    fix_length: bool = True

    def __call__(self, audio: np.ndarray):
        if not self.caching:
            self.aug_index = random.choice(range(0, len(self.factors)))
        factor = self.factors[self.aug_index]
        stretched = librosa.effects.time_stretch(audio, rate=factor)
        if self.fix_length:
            fix_length = FixLength(length=self.sample_length * self.samplerate)
            stretched = fix_length(stretched)

        return stretched


@dataclass
class RandomPitchShift:
    """Shift the pitch of a waveform by n_steps steps .

    Parameters:
        samplerate (float): sample rate of the sample
        factors (float): range of desired factors for pitch shift
        aug_index (int): index of the chosen factor for pitchshift. It will be randomly chosen from the desired range (if not passed while initilization)
        caching (bool): if we are caching, the DiskCached dataset will dynamically pass copy index of data item to the transform (to set aug_index). Otherwise the aug_index will be chosen randomly in every call of transform

    Args:
        audio (np.ndarray): data sample
    Returns:
        np.ndarray: pitch shifted data sample
    """

    samplerate: float
    factors: list = field(
        default_factory=lambda: list(range(-5, 0)) + list(range(1, 6))
    )

    aug_index: int = 0
    caching: bool = False

    def __call__(self, audio: np.ndarray):
        if not self.caching:
            self.aug_index = random.choice(range(0, len(self.factors)))

        semitone = self.factors[self.aug_index]
        return librosa.effects.pitch_shift(audio, sr=self.samplerate, n_steps=semitone)


@dataclass
class RandomAmplitudeScale:
    """Scales the maximum amplitude of the incoming signal to a random amplitude chosen from a
    range.

    Parameters:
        samplerate (float): sample rate of the sample
        min_amp (float): minimum of the amplitude range in volts
        max_amp (float): maximum of the amplitude range in volts
        factors (float): range of desired factors for amplitude scaling
        aug_index (int): index of the chosen factor for pitchshift. It will be randomly chosen from the desired range (if not passed while initilization)
        caching (bool): if we are caching, the DiskCached dataset will dynamically pass copy index of data item to the transform (to set aug_index). Otherwise the aug_index will be chosen randomly in every call of transform


    Args:
        data (np.ndarray): input (single-or multi-channel) signal.

    Returns:
        np.ndarray: scaled version of the signal.
    """

    samplerate: float
    min_amp: float = 0.058
    max_amp: float = 0.159
    factors: list = field(
        default_factory=lambda: list(range(int(1000 * 0.058), int(1000 * 0.159), 10))
    )
    aug_index: int = 0
    caching: bool = False

    def __post_init__(self):
        # converted to mVolt
        self.factors = list(
            range(int(1000 * self.min_amp), int(1000 * self.max_amp), 10)
        )

    def __call__(self, audio: np.ndarray):
        if not self.caching:
            self.aug_index = random.choice(range(0, len(self.factors)))

        max_amplitude = self.factors[self.aug_index] / 1000
        max_data_amplitude = np.max(np.abs(audio.ravel()))
        if max_data_amplitude == 0:
            max_data_amplitude = 1.0

        return audio / max_data_amplitude * max_amplitude


@dataclass
class AddWhiteNoise:
    """Add white noise to the data sample with a known ratio.

    Parameters:
        samplerate (float): sample rate of the sample
        factors (float): range of desired ratios for added noise
        aug_index (int): index of the chosen factor for noise. It will be randomly chosen from the desired range (if not passed while initilization)
        caching (bool): if we are caching the DiskCached dataset will dynamically pass copy index of data item to the transform (to set aug_index). Otherwise the aug_index will be chosen randomly in every call of transform

    Args:
        audio (np.ndarray): data sample
    Returns:
        np.ndarray: data sample with added noise
    """

    samplerate: float
    factors: list = field(default_factory=lambda: [0.1, 0.2, 0.3, 0.4, 0.5])
    aug_index: int = 0
    caching: bool = False

    def __call__(self, audio: np.ndarray):
        if not self.caching:
            self.aug_index = random.choice(range(0, len(self.factors)))

        noise = np.random.normal(0, audio.std(), audio.size)
        noise_factor = self.factors[self.aug_index]
        noisy_audio = audio + noise_factor * noise
        return noisy_audio


@dataclass
class RIR:
    """Convolves a RIR (room impluse response) to the data sample.

    Parameters:
        samplerate (float): sample rate of the sample
        rir_audio (str): path to a sample room impluse response in the .wav format
        caching (bool): if we are caching the DiskCached dataset will dynamically pass copy index of data item to the transform (to set aug_index). Otherwise the aug_index will be chosen randomly in every call of transform

    Args:
        audio (np.ndarray): data sample
    Returns:
        np.ndarray: data sample convolved with RIR
    """

    samplerate: float
    rir_audio: str
    caching: bool = False

    def __call__(self, audio):
        SAMPLE_RIR = download_asset(self.rir_audio)
        rir_raw, rir_sample_rate = torchaudio.load(SAMPLE_RIR)
        rir = rir_raw[:, int(rir_sample_rate * 1.01) : int(rir_sample_rate * 1.3)]
        rir = rir / torch.norm(rir, p=2)
        RIR = torch.flip(rir, [1])

        t_audio = torch.nn.functional.pad(
            torch.from_numpy(audio), (RIR.shape[1] - 1, 0)
        )
        rir_augmented = torch.nn.functional.conv1d(t_audio[None, ...], RIR[None, ...])[
            0
        ].numpy()

        return rir_augmented
